import org.antlr.runtime.*;

import java.io.InputStream;
import java.io.IOException;
import java.util.List;
import java.util.Stack;

public class Interpreter {
    public InterpreterListener listener = // default response to messages
        new InterpreterListener() {
            public void info(String msg) { System.out.println(msg); }
            public void error(String msg) { System.err.println(msg); }
            public void error(String msg, Exception e) { error(msg); }
            public void error(String msg, Token t) {
                error("line "+t.getLine()+": "+msg);
            }
        };

    public static final ReturnValue sharedReturnValue = new ReturnValue();

    // START: state
    MemorySpace globals = new MemorySpace("globals");        // global memory 
    MemorySpace currentSpace = globals;
    Stack<FunctionSpace> stack = new Stack<FunctionSpace>(); // call stack
    TokenRewriteStream tokens; // token buffer represents the code memory
    PieLexer lex;              // lexer/parser are part of the processor
    PieParser parser;
    // END: state

    public void interp(InputStream input)
        throws RecognitionException, IOException
    {
        lex = new PieLexer(new ANTLRInputStream(input));
        tokens = new TokenRewriteStream(lex);
        parser = new PieParser(tokens, this);            // create parser
        parser.program();                                // parse program
    }

    /** Begin execution at tokenIndex using a particular rule */
    // START: exec
    public Object exec(String rule, int tokenIndex) {
        Object rv = null;
        try {
            int oldPosition = parser.input.index(); // save current location
            parser.input.seek(tokenIndex); // seek to place to start execution
            try { // which rule are we executing?
                if ( rule.equals("expr") ) { rv = parser.expr(false).value; }
                else if ( rule.equals("slist") ) { parser.slist(false); }
                else listener.error("error: can't start at "+rule);
            }
            finally { parser.input.seek(oldPosition); }// restore location
        }
        catch (Exception e) {
            listener.error("can't exec code @ index "+tokenIndex, e);
        }
        return rv;
    }
    // END: exec

    public void aggr(String name, List<Token> fields) {
        StructSymbol s = new StructSymbol(name, fields);
        currentSpace.define(s);
    }

    // START: def
    public void def(String name, List<Token> args, Token codeStart) {
        Symbol f = new FunctionSymbol(name, args, codeStart);
        currentSpace.define(f);
    }
    // END: def

    // START: call
    public Object call(String name, List<Object> args) {
        FunctionSymbol f = (FunctionSymbol)load(name); // find function
        if ( f==null ) { listener.error("no function "+name); return null; }
        FunctionSpace fspace = new FunctionSpace(f);
        MemorySpace saveSpace = currentSpace;
        currentSpace = fspace;

        if ( f.formalArgs==null && args.size()>0 || // args compatible?
             f.formalArgs!=null && f.formalArgs.size()!=args.size() ) {
            listener.error("function "+f.name+" argument list mismatch");
            return null;
        }
        // define parameters according to order in formalArgs
        for (int i = 0; f.formalArgs!=null && i < f.formalArgs.size(); i++) {
            Token a = (Token)f.formalArgs.get(i); // get formal name
            fspace.put(a.getText(), args.get(i)); // set name=arg-value
        }
        Object result = null;
        stack.push(fspace); // Push parameter/local space for this call
        try { exec("slist", f.firstTokenOfBlock.getTokenIndex()); }
        catch (ReturnValue rv) { result = rv.value; } // trap return value
        stack.pop(); // toss out local and parameter space
        currentSpace = saveSpace;
        return result;
    }
    // END: call

    public void ret(Object value) {
        sharedReturnValue.value = value;
        throw sharedReturnValue;
    }

    public void assign(String id, Object value) {
        MemorySpace s = getSpaceWithSymbol(id);
        if ( s==null ) s = currentSpace;
        s.put(id, value);
    }

    public Object load(Token idT) {
        Object v = load(idT.getText());
        if ( v==null ) listener.error("no such variable "+idT.getText(), idT);
        return v;
    }

    public void fieldassign(StructInstance struct, Token idT, Object value) {
        String id = idT.getText();
        if ( struct==null ) return;
        if ( struct.def.fields.get(id) == null ) {
            listener.error("can't assign; "+struct.name+" has no "+id+
                            " field", idT);
            return;
        }
        struct.put(id, value);
    }

    public Object fieldload(StructInstance struct, Token idT) {
        String id = idT.getText();
        if ( struct==null ) return null;
        if ( struct.def.fields.get(id)==null ) {
            listener.error(struct.name+" has no "+id+" field", idT);
            return null;
        }
        return struct.get(id);
    }

    // START: whileloop
    public void whileloop(Token condStart, Token codeStart) {
        Boolean c = (Boolean)exec("expr", condStart.getTokenIndex());
        while ( c ) {
            exec("slist", codeStart.getTokenIndex());
            c = (Boolean)exec("expr", condStart.getTokenIndex());
        }
    }
    // END: whileloop

    // START: ifstat
    public void ifstat(Token condStart, Token codeStart, Token elseCodeStart) {
        // evaluate the condition expression
        Boolean b = (Boolean)exec("expr", condStart.getTokenIndex());
        if ( b ) { // if true, do the then clause
            exec("slist", codeStart.getTokenIndex());
        }
        // do else clause if !b and else exists
        else if ( elseCodeStart!=null ) {
            exec("slist", elseCodeStart.getTokenIndex());            
        }
    }
    // END: ifstat

    public boolean eq(Object a, Object b) { return a.equals(b); }

    public boolean lt(Object a, Object b) {
        if ( a instanceof Number && b instanceof Number ) {
            Number x = (Number)a;
            Number y = (Number)b;
            return x.floatValue() < y.floatValue();
        }
        return false;
    }

    public Object op(Object a, char op, Object b) {
        if ( a instanceof Float || b instanceof Float ) {
            float x = ((Number)a).floatValue();
            float y = ((Number)b).floatValue();
            switch (op) {
                case '+' : return x + y;
                case '-' : return x - y;
                case '*' : return x * y;
            }
        }
        if ( a instanceof Integer || b instanceof Integer ) {
            int x = ((Number)a).intValue();
            int y = ((Number)b).intValue();
            switch (op) {
                case '+' : return x + y;
                case '-' : return x - y;
                case '*' : return x * y;
            }
        }
        return 0;
    }

    public Object add(Object a, Object b) {
        if ( a instanceof String || b instanceof String ) {
            return a.toString() + b.toString();
        }
        return op(a, '+', b);
    }
    public Object sub(Object a, Object b) { return op(a, '-', b); }
    public Object mul(Object a, Object b) { return op(a, '*', b); }

    public Object instance(StructSymbol struct) {
        return new StructInstance(struct);
    }

    public Object instance(String structName) {
        StructSymbol struct = (StructSymbol)load(structName);
        return instance(struct);
    }

    /** In a.b, match a, in a.b.c, match a.b, in a.b.c.d, match a.b.c, etc... */
    public StructInstance getStructSpace(List<Token> names) {
        MemorySpace s = currentSpace;
        for (Token t : names) {
            String name = t.getText();
            Object o = s.get(name);
            if ( !(o instanceof StructInstance) ) {
                // make a good error message:
                int firstToken = names.get(0).getTokenIndex();
                int idPastLastDot = names.get(names.size()-1).getTokenIndex()+2;
                String exprS = tokens.toString(firstToken,idPastLastDot);
                listener.error(name+" is not a struct in "+exprS, t);
                return null;
            }
            s = (StructInstance)s.get(name);
        }
        return (StructInstance)s;
    }

    // START: getSpaceWithSymbol
    /** Return scope holding id's value; current func space or global. */
    public MemorySpace getSpaceWithSymbol(String id) {
        if ( stack.size()>0 && stack.peek().get(id)!=null ) { // in top stack?
            return stack.peek();
        }
        if ( globals.get(id)!=null ) return globals;          // in globals?
        return null;                                          // nowhere
    }
    // END: getSpaceWithSymbol

    // START: load
    /** Find id in current semantic context */
    public Object load(String id) {
        MemorySpace s = getSpaceWithSymbol(id);
        if ( s!=null ) return s.get(id);
        return null;
    }
    // END: load
}
