/** A simple dynamically-typed language that smacks of Python */
grammar Pie;

@members {
Interpreter interp;
public PieParser(TokenStream input, Interpreter interp) {
    this(input);
    this.interp = interp;
}
}

// START: program
program : ( functionDefinition[false] | statement[false] )+ EOF ;
// END: program

// START: structDefinition
structDefinition[boolean defer]
    :   'struct' name=ID '{' fields+=ID (',' fields+=ID)* '}' NL
        {if (!defer) interp.aggr($name.text, $fields);}
    ;
// END: structDefinition
    
// START: functionDefinition
functionDefinition[boolean defer]
    :   'def' name=ID '(' (args+=ID (',' args+=ID)*)? ')' slist[true]
        {if (!defer) interp.def($name.text, $args, $slist.start);}
    ;
// END: functionDefinition
    
slist[boolean defer]
    :   ':' NL statement[defer]+ '.' NL
    |   statement[defer]
    ;

// START: statement
statement[boolean defer]
    :   structDefinition[defer]
    |   assign[defer]
    |   'return' expr[defer] NL  {if (!defer) interp.ret($expr.value);}
    |   'print' expr[defer] NL   {if (!defer) System.out.println($expr.value);}
    |   ifstat[defer]
    |   'while' expr[true] slist[true]
        {if (!defer) interp.whileloop($expr.start, $slist.start);}
    |   call[defer] NL
    |   NL
    ;
// END: statement

// START: ifstat
ifstat[boolean defer]
    :   'if' expr[true] c=slist[true] ('else' el=slist[true])?
        {if (!defer) interp.ifstat($expr.start, $c.start, $el.start);}
    ;
// END: ifstat

// START: assign
assign[boolean defer]
    :   ID '=' expr[defer] NL                        // E.g., "a = 3"
        {if (!defer) interp.assign($ID.text, $expr.value);}
    |   structscope[defer] '.' ID '=' expr[defer] NL // E.g., "a.b.x = 3"
        {if (!defer) interp.fieldassign($structscope.value, $ID, $expr.value);}
    ;
// END: assign

// START: call
call[boolean defer] returns [Object value]
@init { List<Object> args = new ArrayList<Object>(); }
    :   ID '('
        (   a=expr[defer]        {args.add($a.value);}
            ( ',' a=expr[defer]  {args.add($a.value);} )*
        )?
        ')'
        {if (!defer) $value = interp.call($ID.text, args);}
    ;
// END: call

expr[boolean defer] returns [Object value]
    :   a=addexpr[defer]        {if (!defer) $value=$a.value;}
        ('==' b=addexpr[defer]  {if (!defer) $value = interp.eq($a.value, $b.value);}
        |'<'  b=addexpr[defer]  {if (!defer) $value = interp.lt($a.value, $b.value);}
        )?
    ;

// START: addexpr
addexpr[boolean defer] returns [Object value]
    :   a=mulexpr[defer]       {if (!defer) $value=$a.value;}
        ( '+' b=mulexpr[defer]
          {if (!defer) $value=interp.add($a.value, $b.value);}
        | '-' b=mulexpr[defer]
          {if (!defer) $value=interp.sub($a.value, $b.value);}
        )*
    ;
// END: addexpr

mulexpr[boolean defer] returns [Object value]
    :   a=atom[defer]          {$value=$a.value;}
        ('*' b=atom[defer]     {if (!defer) $value=interp.mul($a.value, $b.value);})*
    ;

// START: atom
atom[boolean defer] returns [Object value]
    :   INT     {if (!defer) $value = $INT.int;} // return int value
    |   ID      {if (!defer) $value = interp.load($ID);} // look up value
    // ...
// END: atom
    |   CHAR    {if (!defer) $value = new Character($CHAR.text.charAt(1));}
    |   FLOAT   {if (!defer) $value = new Float($FLOAT.text);}
    |   STRING  {String s = $STRING.text; if (!defer) $value = s.substring(1,s.length()-1);}
    |   structscope[defer] '.' ID // field access
        {if ( !defer ) $value = interp.fieldload($structscope.value, $ID);}
    |   call[defer] {if (!defer) $value = $call.value;}
    |   'new' ID    {if (!defer) $value = interp.instance($ID.text);}
    |   '(' expr[defer] ')'
    ;

// START: structscope
/** return struct scope to left of last '.'; a.b in a.b.c, for example */
structscope[boolean defer] returns [StructInstance value] 
    :   ids+=ID ('.' ids+=ID)*
        {if (!defer) $value = interp.getStructSpace($ids);}
    ;
// END: structscope

// L e x i c a l  R u l e s

NL  :   '\r'? '\n' ;

ID  :   LETTER (LETTER | '0'..'9')*  ;

fragment
LETTER
    :   ('a'..'z' | 'A'..'Z')
    ;

CHAR:   '\'' . '\'' ;

STRING: '\"' .* '\"' ;

INT :   '0'..'9'+ ;
    
FLOAT
    :   INT '.' INT*
    |   '.' INT+
    ;

WS  :   (' '|'\t') {$channel=HIDDEN;} ;

SL_COMMENT
    :   '#' ~('\r'|'\n')* {$channel=HIDDEN;}
    ;
